#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021/08/12
@author: pan xia
多分支卷积神经网络/多尺度(分辨率)特征提取
"""
import tensorflow as tf
import keras.backend as K
import datetime, sys
import numpy as np
from keras.models import Model
from keras.optimizers import Adam
from keras.layers import Conv2D, AveragePooling2D, MaxPooling2D, Dense, Bidirectional, LSTM, GRU, CuDNNLSTM, CuDNNGRU, \
    Permute, GlobalAveragePooling2D
from keras.layers import Input, add, BatchNormalization, Activation, Dropout, Concatenate, Flatten, merge, Reshape, \
    Lambda, SpatialDropout2D
import copy
from keras.callbacks import ReduceLROnPlateau, EarlyStopping
from keras.utils import plot_model
from CNN_Attention import AttentionLayer


class XpModel:

    def Conv2d_BLOCK1(self, x, nb_filter, kernel_size, dropout=None, bn_axis=-1, strides=(1, 1), padding='same',
                     withFiltersChange=False, sampwithTwoStrides=False, Convertto2D=False):
        inpt = x
        x = Conv2D(nb_filter=nb_filter, kernel_size=kernel_size, padding=padding, strides=strides)(x)
        x = BatchNormalization(axis=bn_axis)(x)
        x = Activation('relu')(x)

        x = SpatialDropout2D(rate=0.2)(x)  # SpatialDropout2D

        # Attention here.
        x = AttentionLayer(x, 'se_block')  # cbam_block, se_block

        # x = SpatialDropout2D(rate=0.2)(x)  # SpatialDropout2D

        if withFiltersChange:
            x = add([x, Conv2D(nb_filter=nb_filter, kernel_size=kernel_size, padding=padding, strides=(1, 1))(inpt)])
        elif sampwithTwoStrides:
            x = add([x, Conv2D(nb_filter=nb_filter, kernel_size=kernel_size, padding=padding, strides=(1, 2))(inpt)])
        elif Convertto2D:
            x = add([x, Conv2D(nb_filter=nb_filter, kernel_size=kernel_size, padding=padding, strides=(2, 2))(inpt)])
        else:
            x = add([x, inpt])
        return Activation('relu')(x)

    def get_CNN(self, inputDim, outDimen):  # valid or same

        dim_leads, dim_freq = inputDim
        print(dim_leads)

        inpt = Input(shape=inputDim)
        print(inpt.shape)

        inp = Reshape((dim_leads, dim_freq, 1))(inpt)
        print(inp.shape)

        # 分支1
        x = []
        if dim_leads == 12:
            x = self.Conv2d_BLOCK1(inp, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch1_fusion = x # 分支1融合输出(12*250*64)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(5, 3), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(5, 9), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(5, 27), strides=(1, 1))
        elif dim_leads == 6:
            x = self.Conv2d_BLOCK1(inp, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch1_fusion = x # 分支1融合输出(6*250*64)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(3, 3), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(3, 9), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(3, 27), strides=(1, 1))
        elif dim_leads == 4:
            x = self.Conv2d_BLOCK1(inp, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch1_fusion = x # 分支1融合输出(4*250*64)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(2, 3), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(2, 9), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))
        elif dim_leads == 3:
            x = self.Conv2d_BLOCK1(inp, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch1_fusion = x # 分支1融合输出(3*250*64)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 3), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(2, 9), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))
        elif dim_leads == 2:
            x = self.Conv2d_BLOCK1(inp, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch1_fusion = x # 分支1融合输出(2*250*64)
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 3), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(1, 9), strides=(1, 1))
            x = self.Conv2d_BLOCK1(x, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))

        branch1_output = x
        # print(branch1_output.shape)

        # 分支2
        # branch2_scale = AveragePooling2D(pool_size=(1, 2), strides=None, padding='same')(inp)
        branch2_scale = Lambda(lambda m: m[:, :, ::2, :])(inp)  # 对输入进行降2采样

        y = []
        if dim_leads == 12:
            y = self.Conv2d_BLOCK1(branch2_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            y = add([y, branch1_fusion]) # 分支1分支2融合
            y = Activation('relu')(y)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch2_fusion = y # 分支2融合输出(12*125*64)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(5, 3), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(5, 9), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(5, 27), strides=(1, 1))
        elif dim_leads == 6:
            y = self.Conv2d_BLOCK1(branch2_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            y = add([y, branch1_fusion]) # 分支1分支2融合
            y = Activation('relu')(y)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch2_fusion = y # 分支2融合输出(6*125*64)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(3, 3), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(3, 9), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(3, 27), strides=(1, 1))
        elif dim_leads == 4:
            y = self.Conv2d_BLOCK1(branch2_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            y = add([y, branch1_fusion]) # 分支1分支2融合
            y = Activation('relu')(y)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch2_fusion = y # 分支2融合输出(4*125*64)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(2, 3), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(2, 9), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))
        elif dim_leads == 3:
            y = self.Conv2d_BLOCK1(branch2_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            y = add([y, branch1_fusion]) # 分支1分支2融合
            y = Activation('relu')(y)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch2_fusion = y # 分支2融合输出(3*125*64)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 3), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(2, 9), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))
        elif dim_leads == 2:
            y = self.Conv2d_BLOCK1(branch2_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            y = add([y, branch1_fusion]) # 分支1分支2融合
            y = Activation('relu')(y)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            branch2_fusion = y # 分支2融合输出(2*125*64)
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 3), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(1, 9), strides=(1, 1))
            y = self.Conv2d_BLOCK1(y, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))

        branch2_output = y
        # print(branch2_output.shape)

        # 分支3
        # branch3_scale = AveragePooling2D(pool_size=(1, 4), strides=None, padding='same')(inp)
        branch3_scale = Lambda(lambda n: n[:, :, ::4, :])(inp)  # 对输入进行降4采样

        z = []
        if dim_leads == 12:
            z = self.Conv2d_BLOCK1(branch3_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            z = add([z, branch2_fusion]) # 分支2分支3融合
            z = Activation('relu')(z)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(5, 3), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(5, 9), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(5, 27), strides=(1, 1))
        elif dim_leads == 6:
            z = self.Conv2d_BLOCK1(branch3_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            z = add([z, branch2_fusion]) # 分支2分支3融合
            z = Activation('relu')(z)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(3, 3), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(3, 9), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(3, 27), strides=(1, 1))
        elif dim_leads == 4:
            z = self.Conv2d_BLOCK1(branch3_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            z = add([z, branch2_fusion]) # 分支2分支3融合
            z = Activation('relu')(z)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(2, 3), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(2, 9), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))
        elif dim_leads == 3:
            z = self.Conv2d_BLOCK1(branch3_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            z = add([z, branch2_fusion]) # 分支2分支3融合
            z = Activation('relu')(z)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 3), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(2, 9), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))
        elif dim_leads == 2:
            z = self.Conv2d_BLOCK1(branch3_scale, nb_filter=64, kernel_size=(1, 15), strides=(1, 2), sampwithTwoStrides=True)
            z = add([z, branch2_fusion]) # 分支2分支3融合
            z = Activation('relu')(z)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 9), strides=(1, 2), sampwithTwoStrides=True)
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 3), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(1, 9), strides=(1, 1))
            z = self.Conv2d_BLOCK1(z, nb_filter=64, kernel_size=(2, 27), strides=(1, 1))

        branch3_output = z
        # print(branch3_output.shape)

        print("shape before flatten or GlobalAveragePooling2D--", np.shape(branch1_output))
        branch1_output = GlobalAveragePooling2D()(branch1_output)
        print("shape after flatten or GlobalAveragePooling2D--", np.shape(branch1_output))
        print("shape before flatten or GlobalAveragePooling2D--", np.shape(branch2_output))
        branch2_output = GlobalAveragePooling2D()(branch2_output)
        print("shape after flatten or GlobalAveragePooling2D--", np.shape(branch2_output))
        print("shape before flatten or GlobalAveragePooling2D--", np.shape(branch3_output))
        branch3_output = GlobalAveragePooling2D()(branch3_output)
        print("shape after flatten or GlobalAveragePooling2D--", np.shape(branch3_output))

        # 分支拼接
        x = Concatenate(axis=1)([branch1_output, branch2_output, branch3_output])
        print(x.shape)

        # x = Dense(units=64)(x)
        # x = BatchNormalization(axis=-1)(x)
        # x = Activation('relu')(x)
        # x = Dropout(0.3)(x)
        # print(x.shape)

        outputs = Dense(units=outDimen, activation='sigmoid')(x)
        print(outputs.shape)
        md = Model(inputs=inpt, outputs=outputs)
        return md

    def getModel(self, data_shape, loss, outDimen):
        dim_leads, dim_freq = int(data_shape[1]), int(data_shape[2])
        model = self.get_CNN([dim_leads, dim_freq], outDimen)  # Get model
        model.compile(optimizer=Adam(), loss=loss)
        return model

    def getClassWeight(self, y):
        tmp = np.sum(y, axis=0) / len(list(y))
        # tmp = 1.0 / tmp
        tmp = np.log2(1.0 / tmp)
        return tmp

    def custom_loss(self, class_weight):
        print("Class_weight for custom_loss", class_weight)

        def getLoss(yTrue, yPred):
            tmp = K.binary_crossentropy(yTrue, yPred) * class_weight
            #         return tmp
            return K.mean(tmp, axis=-1)

        return getLoss

    def binary_focal_loss(self, class_weight, gamma=2, alpha=.25):
        """
        Binary form of focal loss.
        focal_loss(p_t) = -alpha_t * (1 - p_t)**gamma * log(p_t)
        Usage:
         model.compile(loss=binary_focal_loss(alpha=.25, gamma=2))
        """
        alpha = tf.constant(alpha, dtype=tf.float32)
        gamma = tf.constant(gamma, dtype=tf.float32)

        def binary_focal_loss_fixed(y_true, y_pred):
            """
            y_true shape need be (None,1)
            y_pred need be compute after sigmoid
            """
            #             y_true = tf.cast(y_true, tf.float32)
            alpha_t = alpha  # 此时，如果gamma=0, alpha=1，则完全等价于binary_crossentropy，有些版本则是下面这样的
            #         alpha_t = y_true*alpha + (K.ones_like(y_true)-y_true)*(1-alpha)

            p_t = y_true * y_pred + (K.ones_like(y_true) - y_true) * (K.ones_like(y_true) - y_pred) + K.epsilon()
            focal_loss = - alpha_t * K.pow((K.ones_like(y_true) - p_t), gamma) * K.log(p_t)
            #         return focal_loss
            return K.mean(focal_loss)

        return binary_focal_loss_fixed

    def startTrain(self, x_train, y_train, outDimen, loss='mse', epochs=100, useDataAugment=True):

        early_stop = EarlyStopping(monitor='val_loss', patience=8, verbose=2, mode='min')
        learning_rate_reduction = ReduceLROnPlateau(monitor='val_loss', patience=3, verbose=2, factor=0.5,
                                                    min_lr=0.00001)

        self.model = self.getModel(data_shape=x_train.shape, loss=self.custom_loss(self.getClassWeight(y_train)),
                                   outDimen=outDimen)
        self.model.fit(x_train, y_train, epochs=epochs, batch_size=96, verbose=2,
                       shuffle=True, validation_split=0.1,
                       callbacks=[learning_rate_reduction])  # callbacks=[early_stop,learning_rate_reduction]

        return self.model

    def obtainPredictions(self, x_test, y_test, threshold):
        predProbs = self.model.predict(x_test)
        predLabels = np.array(predProbs > threshold, dtype='int')
        score = self.model.evaluate(x_test, y_test, verbose=0)
        return predProbs, predLabels, score

